@@ -2,7 +2,7 @@
 
 Copyright (c) 1995, 2017, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2009, Google Inc.
-Copyright (c) 2014, 2020, MariaDB Corporation.
+Copyright (c) 2014, 2019, MariaDB Corporation.
 
 Portions of this file contain modifications contributed and copyrighted by
 Google, Inc. Those modifications are gratefully acknowledged and are described
@@ -83,9 +83,12 @@ reduce the size of the log.
 /** Redo log system */
 log_t	log_sys;
 
-/** Whether to require checksums on the redo log pages */
+/** Whether to generate and require checksums on the redo log pages */
 my_bool	innodb_log_checksums;
 
+/** Pointer to the log checksum calculation function */
+log_checksum_func_t log_checksum_algorithm_ptr;
+
 /* Next log block number to do dummy record filling if no log records written
 for a while */
 static ulint		next_lbn_to_pad = 0;
@@ -159,19 +162,16 @@ log_buf_pool_get_oldest_modification(void)
 void log_buffer_extend(ulong len)
 {
 	const ulong new_buf_size = ut_calc_align(len, srv_page_size);
-	byte* new_buf = static_cast<byte*>(ut_malloc_dontdump(new_buf_size));
-	TRASH_ALLOC(new_buf, new_buf_size);
-	byte* new_flush_buf =
-		static_cast<byte*>(ut_malloc_dontdump(new_buf_size));
-	TRASH_ALLOC(new_flush_buf, new_buf_size);
+	byte* new_buf = static_cast<byte*>(
+		ut_malloc_dontdump(new_buf_size * 2));
+	TRASH_ALLOC(new_buf, new_buf_size * 2);
 
 	log_mutex_enter();
 
 	if (len <= srv_log_buffer_size) {
 		/* Already extended enough by the others */
 		log_mutex_exit();
-		ut_free_dodump(new_buf, new_buf_size);
-		ut_free_dodump(new_flush_buf, new_buf_size);
+		ut_free_dodump(new_buf, new_buf_size * 2);
 		return;
 	}
 
@@ -179,14 +179,14 @@ void log_buffer_extend(ulong len)
 		" exceeds innodb_log_buffer_size="
 		<< srv_log_buffer_size << " / 2). Trying to extend it.";
 
-	byte* old_buf = log_sys.buf;
-	byte* old_flush_buf = log_sys.flush_buf;
+	const byte* old_buf_begin = log_sys.buf;
 	const ulong old_buf_size = srv_log_buffer_size;
-
+	byte* old_buf = log_sys.first_in_use
+		? log_sys.buf : log_sys.buf - old_buf_size;
 	srv_log_buffer_size = new_buf_size;
 	log_sys.buf = new_buf;
-	log_sys.flush_buf = new_flush_buf;
-	memcpy(new_buf, old_buf, log_sys.buf_free);
+	log_sys.first_in_use = true;
+	memcpy(log_sys.buf, old_buf_begin, log_sys.buf_free);
 
 	log_sys.max_buf_free = new_buf_size / LOG_BUF_FLUSH_RATIO
 		- LOG_BUF_FLUSH_MARGIN;
@@ -194,7 +194,6 @@ void log_buffer_extend(ulong len)
 	log_mutex_exit();
 
 	ut_free_dodump(old_buf, old_buf_size);
-	ut_free_dodump(old_flush_buf, old_buf_size);
 
 	ib::info() << "innodb_log_buffer_size was extended to "
 		<< new_buf_size << ".";
@@ -211,9 +210,9 @@ log_calculate_actual_len(
 {
 	ut_ad(log_mutex_own());
 
+	const ulint	framing_size = log_sys.framing_size();
 	/* actual length stored per block */
-	const ulint	len_per_blk = OS_FILE_LOG_BLOCK_SIZE
-		- (LOG_BLOCK_HDR_SIZE + LOG_BLOCK_TRL_SIZE);
+	const ulint	len_per_blk = OS_FILE_LOG_BLOCK_SIZE - framing_size;
 
 	/* actual data length in last block already written */
 	ulint	extra_len = (log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE);
@@ -222,8 +221,7 @@ log_calculate_actual_len(
 	extra_len -= LOG_BLOCK_HDR_SIZE;
 
 	/* total extra length for block header and trailer */
-	extra_len = ((len + extra_len) / len_per_blk)
-		* (LOG_BLOCK_HDR_SIZE + LOG_BLOCK_TRL_SIZE);
+	extra_len = ((len + extra_len) / len_per_blk) * framing_size;
 
 	return(len + extra_len);
 }
@@ -341,26 +339,24 @@ log_write_low(
 	ulint		str_len)	/*!< in: string length */
 {
 	ulint	len;
-	ulint	data_len;
-	byte*	log_block;
 
 	ut_ad(log_mutex_own());
+	const ulint trailer_offset = log_sys.trailer_offset();
 part_loop:
 	/* Calculate a part length */
 
-	data_len = (log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE) + str_len;
+	ulint data_len = (log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE) + str_len;
 
-	if (data_len <= OS_FILE_LOG_BLOCK_SIZE - LOG_BLOCK_TRL_SIZE) {
+	if (data_len <= trailer_offset) {
 
 		/* The string fits within the current log block */
 
 		len = str_len;
 	} else {
-		data_len = OS_FILE_LOG_BLOCK_SIZE - LOG_BLOCK_TRL_SIZE;
+		data_len = trailer_offset;
 
-		len = OS_FILE_LOG_BLOCK_SIZE
-			- (log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE)
-			- LOG_BLOCK_TRL_SIZE;
+		len = trailer_offset
+			- log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE;
 	}
 
 	memcpy(log_sys.buf + log_sys.buf_free, str, len);
@@ -368,18 +364,18 @@ part_loop:
 	str_len -= len;
 	str = str + len;
 
-	log_block = static_cast<byte*>(
+	byte* log_block = static_cast<byte*>(
 		ut_align_down(log_sys.buf + log_sys.buf_free,
 			      OS_FILE_LOG_BLOCK_SIZE));
 
 	log_block_set_data_len(log_block, data_len);
 
-	if (data_len == OS_FILE_LOG_BLOCK_SIZE - LOG_BLOCK_TRL_SIZE) {
+	if (data_len == trailer_offset) {
 		/* This block became full */
 		log_block_set_data_len(log_block, OS_FILE_LOG_BLOCK_SIZE);
 		log_block_set_checkpoint_no(log_block,
 					    log_sys.next_checkpoint_no);
-		len += LOG_BLOCK_HDR_SIZE + LOG_BLOCK_TRL_SIZE;
+		len += log_sys.framing_size();
 
 		log_sys.lsn += len;
 
@@ -548,10 +544,10 @@ void log_t::create()
   ut_ad(srv_log_buffer_size >= 16 * OS_FILE_LOG_BLOCK_SIZE);
   ut_ad(srv_log_buffer_size >= 4U << srv_page_size_shift);
 
-  buf= static_cast<byte*>(ut_malloc_dontdump(srv_log_buffer_size));
-  TRASH_ALLOC(buf, srv_log_buffer_size);
-  flush_buf= static_cast<byte*>(ut_malloc_dontdump(srv_log_buffer_size));
-  TRASH_ALLOC(flush_buf, srv_log_buffer_size);
+  buf= static_cast<byte*>(ut_malloc_dontdump(srv_log_buffer_size * 2));
+  TRASH_ALLOC(buf, srv_log_buffer_size * 2);
+
+  first_in_use= true;
 
   max_buf_free= srv_log_buffer_size / LOG_BUF_FLUSH_RATIO -
     LOG_BUF_FLUSH_MARGIN;
@@ -606,8 +602,7 @@ void log_t::files::create(ulint n_files)
 
   this->n_files= n_files;
   format= srv_encrypt_log
-    ? LOG_HEADER_FORMAT_CURRENT | LOG_HEADER_FORMAT_ENCRYPTED
-    : LOG_HEADER_FORMAT_CURRENT;
+    ? LOG_HEADER_FORMAT_ENC_10_4 : LOG_HEADER_FORMAT_10_4;
   subformat= 2;
   file_size= srv_log_file_size;
   lsn= LOG_START_LSN;
@@ -629,8 +624,8 @@ log_file_header_flush(
 	ut_ad(log_write_mutex_own());
 	ut_ad(!recv_no_log_write);
 	ut_a(nth_file < log_sys.log.n_files);
-	ut_ad((log_sys.log.format & ~LOG_HEADER_FORMAT_ENCRYPTED)
-	      == LOG_HEADER_FORMAT_CURRENT);
+	ut_ad(log_sys.log.format == LOG_HEADER_FORMAT_10_4
+	      || log_sys.log.format == LOG_HEADER_FORMAT_ENC_10_4);
 
 	// man 2 open suggests this buffer to be aligned by 512 for O_DIRECT
 	MY_ALIGNED(OS_FILE_LOG_BLOCK_SIZE)
@@ -659,7 +654,7 @@ log_file_header_flush(
 
 	fil_io(IORequestLogWrite, true,
 	       page_id_t(SRV_LOG_SPACE_FIRST_ID, page_no),
-	       univ_page_size,
+	       0,
 	       ulint(dest_offset & (srv_page_size - 1)),
 	       OS_FILE_LOG_BLOCK_SIZE, buf, NULL);
 
@@ -676,7 +671,7 @@ log_block_store_checksum(
 /*=====================*/
 	byte*	block)	/*!< in/out: pointer to a log block */
 {
-	log_block_set_checksum(block, log_block_calc_checksum_crc32(block));
+	log_block_set_checksum(block, log_block_calc_checksum(block));
 }
 
 /******************************************************//**
@@ -779,7 +774,7 @@ loop:
 
 	fil_io(IORequestLogWrite, true,
 	       page_id_t(SRV_LOG_SPACE_FIRST_ID, page_no),
-	       univ_page_size,
+	       0,
 	       ulint(next_offset & (srv_page_size - 1)), write_len, buf, NULL);
 
 	srv_stats.os_log_pending_writes.dec();
@@ -825,8 +820,8 @@ log_write_flush_to_disk_low()
 	os_event_set(log_sys.flush_event);
 }
 
-/** Swap log buffers, and copy the content of last block
-from old buf to the head of the new buf. Thus, buf_free and
+/** Switch the log buffer in use, and copy the content of last block
+from old log buffer to the head of the to be used one. Thus, buf_free and
 buf_next_to_write would be changed accordingly */
 static inline
 void
@@ -835,15 +830,26 @@ log_buffer_switch()
 	ut_ad(log_mutex_own());
 	ut_ad(log_write_mutex_own());
 
+	const byte*	old_buf = log_sys.buf;
 	ulong		area_end = ut_calc_align(
 		log_sys.buf_free, ulong(OS_FILE_LOG_BLOCK_SIZE));
 
-	/* Copy the last block to new buf */
-	ut_memcpy(log_sys.flush_buf,
-		  log_sys.buf + area_end - OS_FILE_LOG_BLOCK_SIZE,
-		  OS_FILE_LOG_BLOCK_SIZE);
+	if (log_sys.first_in_use) {
+		log_sys.first_in_use = false;
+		ut_ad(log_sys.buf == ut_align(log_sys.buf,
+					       OS_FILE_LOG_BLOCK_SIZE));
+		log_sys.buf += srv_log_buffer_size;
+	} else {
+		log_sys.first_in_use = true;
+		log_sys.buf -= srv_log_buffer_size;
+		ut_ad(log_sys.buf == ut_align(log_sys.buf,
+					       OS_FILE_LOG_BLOCK_SIZE));
+	}
 
-	std::swap(log_sys.buf, log_sys.flush_buf);
+	/* Copy the last block to new buf */
+	ut_memcpy(log_sys.buf,
+		  old_buf + area_end - OS_FILE_LOG_BLOCK_SIZE,
+		  OS_FILE_LOG_BLOCK_SIZE);
 
 	log_sys.buf_free %= OS_FILE_LOG_BLOCK_SIZE;
 	log_sys.buf_next_to_write = log_sys.buf_free;
@@ -855,11 +861,9 @@ wait and check if an already running write is covering the request.
 @param[in]	lsn		log sequence number that should be
 included in the redo log file write
 @param[in]	flush_to_disk	whether the written log should also
-be flushed to the file system */
-void
-log_write_up_to(
-	lsn_t	lsn,
-	bool	flush_to_disk)
+be flushed to the file system
+@param[in]	rotate_key	whether to rotate the encryption key */
+void log_write_up_to(lsn_t lsn, bool flush_to_disk, bool rotate_key)
 {
 #ifdef UNIV_DEBUG
 	ulint		loop_count	= 0;
@@ -868,6 +872,7 @@ log_write_up_to(
 	lsn_t           write_lsn;
 
 	ut_ad(!srv_read_only_mode);
+	ut_ad(!rotate_key || flush_to_disk);
 
 	if (recv_no_ibuf_operations) {
 		/* Recovery is running and no operations on the log files are
@@ -1003,7 +1008,7 @@ loop:
 		}
 	}
 
-	if (UNIV_UNLIKELY(srv_shutdown_state > SRV_SHUTDOWN_INITIATED)) {
+	if (UNIV_UNLIKELY(srv_shutdown_state != SRV_SHUTDOWN_NONE)) {
 		service_manager_extend_timeout(INNODB_EXTEND_TIMEOUT_INTERVAL,
 					       "InnoDB log write: "
 					       LSN_PF "," LSN_PF,
@@ -1012,7 +1017,8 @@ loop:
 
 	if (log_sys.is_encrypted()) {
 		log_crypt(write_buf + area_start, log_sys.write_lsn,
-			  area_end - area_start);
+			  area_end - area_start,
+			  rotate_key ? LOG_ENCRYPT_ROTATE_KEY : LOG_ENCRYPT);
 	}
 
 	/* Do the write to the log files */
@@ -1223,7 +1229,7 @@ log_group_checkpoint(lsn_t end_lsn)
 	ut_ad(end_lsn == 0 || end_lsn >= log_sys.next_checkpoint_lsn);
 	ut_ad(end_lsn <= log_sys.lsn);
 	ut_ad(end_lsn + SIZE_OF_MLOG_CHECKPOINT <= log_sys.lsn
-	      || srv_shutdown_state > SRV_SHUTDOWN_INITIATED);
+	      || srv_shutdown_state != SRV_SHUTDOWN_NONE);
 
 	DBUG_PRINT("ib_log", ("checkpoint " UINT64PF " at " LSN_PF
 			      " written",
@@ -1267,7 +1273,7 @@ log_group_checkpoint(lsn_t end_lsn)
 
 	fil_io(IORequestLogWrite, false,
 	       page_id_t(SRV_LOG_SPACE_FIRST_ID, 0),
-	       univ_page_size,
+	       0,
 	       (log_sys.next_checkpoint_no & 1)
 	       ? LOG_CHECKPOINT_2 : LOG_CHECKPOINT_1,
 	       OS_FILE_LOG_BLOCK_SIZE,
@@ -1287,7 +1293,7 @@ void log_header_read(ulint header)
 	fil_io(IORequestLogRead, true,
 	       page_id_t(SRV_LOG_SPACE_FIRST_ID,
 			 header >> srv_page_size_shift),
-	       univ_page_size, header & (srv_page_size - 1),
+	       0, header & (srv_page_size - 1),
 	       OS_FILE_LOG_BLOCK_SIZE, log_sys.checkpoint_buf, NULL);
 }
 
@@ -1334,7 +1340,7 @@ log_append_on_checkpoint(
 /** Make a checkpoint. Note that this function does not flush dirty
 blocks from the buffer pool: it only checks what is lsn of the oldest
 modification in the pool, and writes information about the lsn in
-log files. Use log_make_checkpoint() to flush also the pool.
+log files. Use log_make_checkpoint_at() to flush also the pool.
 @param[in]	sync		whether to wait for the write to complete
 @return true if success, false if a checkpoint write was already running */
 bool log_checkpoint(bool sync)
@@ -1383,7 +1389,7 @@ bool log_checkpoint(bool sync)
 	if (oldest_lsn
 	    > log_sys.last_checkpoint_lsn + SIZE_OF_MLOG_CHECKPOINT) {
 		/* Some log has been written since the previous checkpoint. */
-	} else if (srv_shutdown_state > SRV_SHUTDOWN_INITIATED) {
+	} else if (srv_shutdown_state != SRV_SHUTDOWN_NONE) {
 		/* MariaDB 10.3 startup expects the redo log file to be
 		logically empty (not even containing a MLOG_CHECKPOINT record)
 		after a clean shutdown. Perform an extra checkpoint at
@@ -1408,7 +1414,7 @@ bool log_checkpoint(bool sync)
 	lsn_t		flush_lsn	= oldest_lsn;
 	const lsn_t	end_lsn		= log_sys.lsn;
 	const bool	do_write
-		= srv_shutdown_state <= SRV_SHUTDOWN_INITIATED
+		= srv_shutdown_state == SRV_SHUTDOWN_NONE
 		|| flush_lsn != end_lsn;
 
 	if (fil_names_clear(flush_lsn, do_write)) {
@@ -1418,7 +1424,7 @@ bool log_checkpoint(bool sync)
 
 	log_mutex_exit();
 
-	log_write_up_to(flush_lsn, true);
+	log_write_up_to(flush_lsn, true, true);
 
 	log_mutex_enter();
 
@@ -1450,12 +1456,14 @@ bool log_checkpoint(bool sync)
 	return(true);
 }
 
-/** Make a checkpoint */
-void log_make_checkpoint()
+/** Make a checkpoint at or after a specified LSN.
+@param[in]	lsn		the log sequence number, or LSN_MAX
+for the latest LSN */
+void log_make_checkpoint_at(lsn_t lsn)
 {
 	/* Preflush pages synchronously */
 
-	while (!log_preflush_pool_modified_pages(LSN_MAX)) {
+	while (!log_preflush_pool_modified_pages(lsn)) {
 		/* Flush as much as we can */
 	}
 
@@ -1604,11 +1612,11 @@ loop:
 		} else {
 			ut_ad(!srv_dict_stats_thread_active);
 		}
-		if (recv_sys && recv_sys->flush_start) {
+		if (recv_sys.flush_start) {
 			/* This is in case recv_writer_thread was never
 			started, or buf_flush_page_cleaner_coordinator
 			failed to notice its termination. */
-			os_event_set(recv_sys->flush_start);
+			os_event_set(recv_sys.flush_start);
 		}
 	}
 #define COUNT_INTERVAL 600U
@@ -1673,7 +1681,7 @@ wait_suspend_loop:
 			"Waiting for %s to exit", thread_name);
 		if (srv_print_verbose_log && count > COUNT_INTERVAL) {
 			ib::info() << "Waiting for " << thread_name
-				   << " to exit";
+				   << "to exit";
 			count = 0;
 		}
 		goto loop;
@@ -1716,9 +1724,6 @@ wait_suspend_loop:
 				"Waiting for page cleaner");
 			ib::info() << "Waiting for page_cleaner to "
 				"finish flushing of buffer pool";
-			/* This is a workaround to avoid the InnoDB hang
-			when OS datetime changed backwards */
-			os_event_set(buf_flush_event);
 			count = 0;
 		}
 	}
@@ -1791,7 +1796,7 @@ wait_suspend_loop:
 	if (!srv_read_only_mode) {
 		service_manager_extend_timeout(INNODB_EXTEND_TIMEOUT_INTERVAL,
 			"ensuring dirty buffer pool are written to log");
-		log_make_checkpoint();
+		log_make_checkpoint_at(LSN_MAX);
 
 		log_mutex_enter();
 
@@ -1935,10 +1940,10 @@ void log_t::close()
   m_initialised = false;
   log.close();
 
-  ut_free_dodump(buf, srv_log_buffer_size);
+  if (!first_in_use)
+    buf -= srv_log_buffer_size;
+  ut_free_dodump(buf, srv_log_buffer_size * 2);
   buf = NULL;
-  ut_free_dodump(flush_buf, srv_log_buffer_size);
-  flush_buf = NULL;
 
   os_event_destroy(flush_event);
   rw_lock_free(&checkpoint_lock);
@@ -1949,7 +1954,7 @@ void log_t::close()
   if (!srv_read_only_mode && srv_scrub_log)
     os_event_destroy(log_scrub_event);
 
-  recv_sys_close();
+  recv_sys.close();
 }
 
 /******************************************************//**
@@ -1969,13 +1974,9 @@ log_pad_current_log_block(void)
 	/* We retrieve lsn only because otherwise gcc crashed on HP-UX */
 	lsn = log_reserve_and_open(OS_FILE_LOG_BLOCK_SIZE);
 
-	pad_length = OS_FILE_LOG_BLOCK_SIZE
-		- (log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE)
-		- LOG_BLOCK_TRL_SIZE;
-	if (pad_length
-	    == (OS_FILE_LOG_BLOCK_SIZE - LOG_BLOCK_HDR_SIZE
-		- LOG_BLOCK_TRL_SIZE)) {
-
+	pad_length = log_sys.trailer_offset()
+		- log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE;
+	if (pad_length == log_sys.payload_size()) {
 		pad_length = 0;
 	}
 
